import os
import tempfile
from pathlib import Path
from seeqret import seeqret_init


def test_init():
    with tempfile.TemporaryDirectory() as temp_dir:
        seeqret_init.secrets_init(Path(temp_dir), 'test', 'test@example.com', None)
        # seeqret_init.secrets_init()
        assert os.path.exists(os.path.join(temp_dir, 'seeqret'))
        assert os.path.exists(os.path.join(temp_dir, 'seeqret', 'private.key'))
        assert os.path.exists(os.path.join(temp_dir, 'seeqret', 'public.key'))
        assert os.path.exists(os.path.join(temp_dir, 'seeqret', 'seeqret.key'))
        assert os.path.exists(os.path.join(temp_dir, 'seeqret', 'seeqrets.db'))



import sqlite3
# import click
from seeqret.migrations.utils import current_version
from .. import cd

from . import (
    db_v_001,
    db_v_002,
)


def init_db(vault_dir, user, email):
    with cd(vault_dir):
        cn = sqlite3.connect('seeqrets.db')
        with cn:
            cur = current_version(cn)
            if cur < 1:
                db_v_001.init_db_v_001(vault_dir, user, email)
            if cur < 2:
                db_v_002.init_db_v_002(vault_dir)
        cn.close()



from ..models import Secret, User


class Storage:
    """Base class for storage backends.
    """
    def __init__(self, name, version=None):
        self.name = name
        self.version = version

    def add_user(self, user: User):
        raise NotImplementedError

    def fetch_users(self, **filters):
        raise NotImplementedError   # pragma: no cover

    def fetch_admin(self):
        raise NotImplementedError   # pragma: no cover

    def add_secret(self, app, env, key, value, type='str'):
        raise NotImplementedError   # pragma: no cover

    def add_secret_obj(self, secret: Secret):
        self.add_secret(
            secret.app,
            secret.env,
            secret.key,
            secret.value,
            secret.type,
        )

    def fetch_secrets(self, **filters):
        raise NotImplementedError   # pragma: no cover

    def remove_secrets(self, **filters):
        raise NotImplementedError



from .command_serializer import CommandSerializer       # noqa
from .jsoncrypt_serializer import JsonCryptSerializer   # noqa
from .insecure_json import InsecureJsonSerializer       # noqa
from .env_serializer import EnvSerializer               # noqa



from seeqret.filterspec import FilterSpec


def test_filterspec():
    items = [
        ('*', '*', 'PG_PASSWORD'),
        ('myapp', '*', 'PG_PASSWORD'),
        ('myapp', 'dev', 'PG_PASSWORD'),
    ]


    f = FilterSpec('*')
    assert str(f) == '*:*:*'
    assert list(f.filter(items)) == items

    f = FilterSpec('PG*')
    assert str(f) == "*:*:PG*"
    assert list(f.filter(items)) == items

    f = FilterSpec('myapp:*')
    assert str(f) == 'myapp:*:*'
    assert list(f.filter(items)) == items[1:]

    f = FilterSpec('myapp:dev:*')
    assert str(f) == 'myapp:dev:*'
    assert list(f.filter(items)) == [items[2]]



import traceback


def print_result(res):
    print('\n--RESULT:-------------------')
    print("EXIT:CODE:", res.exit_code)
    print("OUTPUT:", res.output)

    if res.exc_info and res.exception:
        traceback.print_exception(*res.exc_info)
        # print(res.exception)
    # print traceback

    # print("DIR:", dir(res))
    # traceback.print_exc(res.exc_info)
    print('--/RESULT:-------------------')



# from seeqret.seeqret_add import *
# import pytest
#
#
# def test_fetch_pubkey_from_url():
#     pubkey = fetch_pubkey_from_url('https://raw.githubusercontent.com/tkbeorg/tkbe/refs/heads/main/public.key')
#     assert pubkey == 'ilxSnX9+NrwmeIzOFtWrl0lPPkxTEATmC39BILX6rWk=\n'
#
#     with pytest.raises(RuntimeError):
#         fetch_pubkey_from_url('https://raw.githubusercontent.com/tkbeorg/tkbe/refs/heads/main/xxx')



import re
import logging

logger = logging.getLogger(__name__)


def glob_to_regex(glob_pattern: str) -> str:
    """
    Convert a glob pattern to a regex pattern.

    Parameters:
        glob_pattern (str): The glob pattern to convert.

    Returns:
        str: The equivalent regex pattern.
    """
    # Escape all special characters except for glob-specific ones (*, ?, [])
    regex = re.escape(glob_pattern)

    # Convert glob-specific characters
    # '*' matches zero or more characters
    regex = regex.replace(r'\*', '.*')
    # '?' matches exactly one character
    regex = regex.replace(r'\?', '.')
    # '[!...]' matches any character not in the brackets
    regex = regex.replace(r'\[!', '[^')
    # Allow square brackets for ranges
    regex = regex.replace(r'\[', '[').replace(r'\]', ']')

    # Ensure the regex matches the full string
    return f"^{regex}$"


class FilterSpec:
    def __init__(self, filterspec):
        self.filterspec = filterspec
        parts = filterspec.split(':')
        # print("PARTS:", parts)

        if len(parts) == 1:
            self.app = '*'
            self.env = '*'
            self.name = parts[0] or '*'
        if len(parts) == 2:
            self.app = parts[0] or '*'
            self.env = parts[1] or '*'
            self.name = '*'
        if len(parts) == 3:
            self.app = parts[0] or '*'
            self.env = parts[1] or '*'
            self.name = parts[2] or '*'

        logger.debug("FILTERSPEC: %s == %s", filterspec, str(self))

    def to_filterdict(self):
        return dict(
            app=self.app,
            env=self.env,
            key=self.name,
        )

    def __str__(self):
        return self.app + ':' + self.env + ':' + self.name

    def pattern_to_regex(self, pattern: str):
        return re.compile(glob_to_regex(pattern))

    def match_item(self, val: str, pattern: str) -> bool:
        # print("MATCH:ITEM:", val, pattern, end=' ')
        if pattern == '*':
            # print("TRUE")
            return True
        regex = self.pattern_to_regex(pattern)
        # print("REGEX:", regex)
        # print("MATCH:", regex.match(val))
        return regex.match(val) is not None

    def matches(self, item: (str, str, str)) -> bool:
        return all(self.match_item(val, pattern)
                   for val, pattern
                   in zip(item, (self.app, self.env, self.name)))

    def filter(self, items: list[(str, str, str)]) -> list:
        for item in items:
            # print("ITEM:", item)
            # print("-------------------")
            if self.matches(item):
                yield item



from .serializer import serializer, BaseSerializer


@serializer
class EnvSerializer(BaseSerializer):
    """Export to .env file.
    """
    version = 1
    tag = 'env'

    def dumps(self, secrets, system):
        if system == 'linux':
            return "don't write .env files on linux"

        res = []
        for secret in secrets:
            res.append(f'{secret.key}={secret.value}')

        return '\n'.join(res)

    def loads(self, secrets, system):
        raise NotImplementedError()



import os

import click
from .filterspec import FilterSpec
from .models import Secret
from .storage.sqlite_storage import SqliteStorage
from .run_utils import cd


@click.command()
@click.pass_context
@click.argument('name')
@click.argument('value')
@click.option('--app', default='*', show_default=True,
              help='The app to add the secret to')
@click.option('--env', default='*', show_default=True,
              help='The env(ironment) to add the secret to (e.g. dev/prod)')
def key(ctx, name: str, value: str, app: str = None, env: str = None):  # noqa: F811
    """Add a new NAME -> VALUE mapping.

       You can (should) specify the app and environment properties when adding
       a new mapping.
    """
    if ':' in name or ':' in app or ':' in env:
        ctx.fail(click.style(
            'Colon `:` is not valid in key, app, or env', fg='red'
        ))
    click.secho(
        f'Adding a new key: {name}, value: {value}, app: {app}, env: {env}',
        fg='blue'
    )

    with cd(os.environ['SEEQRET']):
        storage = SqliteStorage()
        fspec = FilterSpec(f'{app}:{env}:{name}')
        secrets = storage.fetch_secrets(**fspec.to_filterdict())
        if secrets:
            ctx.fail(click.style(
                f'Secret {", ".join(s.key for s in secrets)} already exists!',
                fg='red'
            ))
        secret = Secret(
            app=app,
            env=env,
            key=name,
            plaintext_value=value,
            type='str'
        )
        storage.add_secret(secret)

        # verify that it worked
        secrets = storage.fetch_secrets(app=app, env=env, key=name)
        if secrets:
            click.secho(
                f'..successfully added: {app}:{env}[{key}]', fg='green'
            )
        else:
            ctx.fail(click.style(
                f'Error: {app}:{env}[{key}] not written to database', fg='red'
            ))



import os
import sqlite3
from os import abort

import click


def fetch_admin(cn):
    admin = cn.execute('''
        select username, email, pubkey
        from users
        where id = 1
    ''').fetchone()
    if not admin:
        click.secho('No admin user', fg='red')
        abort()
    return dict(username=admin[0], email=admin[1], pubkey=admin[2])


def fetch_user(cn, username):
    user = cn.execute('''
        select username, email, pubkey
        from users
        where username = ?
    ''', [username]).fetchone()
    if not user:
        return None
    return dict(username=user[0], email=user[1], pubkey=user[2])


def debug_fetch_users():
    cn = sqlite3.connect(os.path.join(os.environ['SEEQRET'], 'seeqrets.db'))
    with cn:
        return [rec[0] for rec in cn.execute("""
            select username
            from users
            order by username
        """)]


def debug_secrets():
    cn = sqlite3.connect(os.path.join(os.environ['SEEQRET'], 'seeqrets.db'))
    with cn:
        return [rec for rec in cn.execute("""
            select app, env, key, value
            from secrets
            order by app, env, key
        """)]



import json
import os
from pathlib import Path

import click
from click import Context
from .storage.sqlite_storage import SqliteStorage
from . import seeqret_transfer
# from .context import Context
from . import seeqret_init, seeqret_add, cd
from .console_utils import as_table, dochelp
from .fileutils import is_writable
from .filterspec import FilterSpec
from .serializers.serializer import SERIALIZERS
from .cli_group_rm import key as rm_key
from .cli_group_add import key as add_key
import logging

DIRNAME = Path(__file__).parent
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.pass_context
@click.option('-L', '--log', default="ERROR")
def cli(ctx, log):
    logging.basicConfig(level=getattr(logging, log))
    ctx.obj = {
        "seeqrets_dir": os.environ.get("SEEQRETS"),
        "curdir": os.getcwd(),
    }


@cli.command()
def info():
    "debugging"
    with Context(cli) as ctx:
        info = ctx.to_info_dict()
        print(json.dumps(info, indent=4))


@cli.command()
def upgrade():
    """Upgrade the database to the latest version
    """
    with cd(os.environ['SEEQRET']):
        seeqret_init.upgrade_db()


@cli.command()
@click.option('-f', '--filter', default='*', show_default=True,
              help='filterspec (see XXX)')
def list(filter):
    """List the contents of the vault
    """
    with cd(os.environ['SEEQRET']):
        storage = SqliteStorage()
        fspec = FilterSpec(filter)
        as_table("App,Env,Key,Value,Type",
                 storage.fetch_secrets(**fspec.to_filterdict()))


@cli.command()
def users():
    """List the users in the vault
    """
    # print("SEEQRET:DIR:", os.environ['SEEQRET'])
    with cd(os.environ['SEEQRET']):
        storage = SqliteStorage()
        as_table('username,email,publickey',
                 storage.fetch_users())


@cli.command()
def serializers():
    """List available serializers.
    """
    as_table('Name, tag, version, description',
             [(
                 cls.__name__,
                 cls.tag,
                 str(cls.version),
                 dochelp(cls)
             ) for cls in SERIALIZERS.values()])


@cli.command()
def backup():
    serializer = SERIALIZERS['backup']
    with cd(os.environ['SEEQRET']):
        seeqret_transfer.export_secrets(
            'self', FilterSpec('::'),
            serializer, False, False
        )


@cli.command()
@click.pass_context
@click.argument('to')
@click.option('-f', '--filter', default='',
              help='A seeqret filter string (see XXX) ')
@click.option(
    '-s', '--serializer', default='json-crypt',
    help='Name of serializer to use (`seeqret serializers` to list).')
@click.option('-o', '--out', default=None,
              help='Output file (default: stdout).')
@click.option('-w', '--windows', default=False, is_flag=True,
              help='Export to windows format.')
@click.option('-l', '--linux', default=False, is_flag=True,
              help='Export to linux format.')
def export(ctx, to, filter, serializer='json-crypt', out=None,
           windows=False, linux=False):
    """Export the vault to a user
    """
    serializer_cls = SERIALIZERS.get(serializer)
    if not serializer_cls:
        ctx.fail(
            f'Unknown serializer: `{serializer}` ({", ".join(SERIALIZERS.keys())}) '
            '(use `seeqret serializers` to list available serializers).'
        )
    with cd(os.environ['SEEQRET']):
        seeqret_transfer.export_secrets(
            ctx,
            to, FilterSpec(filter),
            serializer_cls, out, windows, linux
        )


@cli.command()
@click.pass_context
@click.option('-u', '--from-user', default='',
              help='Sender.')
@click.option('-f', '--file', default=None,
              help='Path to the vault file to export')
@click.option('-v', '--value', default=None,
              help='(string) value to export for the vault file')
@click.option('-s', '--serializer', default='json-crypt',
              help='Serializer to use (`seeqret serializers` to list).')
def save(ctx, from_user, file, value, serializer):
    """Save exported secrets.
    """
    serializer_cls = SERIALIZERS.get(serializer)
    if not serializer_cls:
        ctx.fail(
            f'Unknown serializer: {serializer} '
            '(use `seeqret serializers` to list available serializers).'
        )
    with cd(os.environ['SEEQRET']):
        seeqret_transfer.import_secrets(
            from_user, file, value, serializer_cls
        )


@cli.command()
@click.pass_context
@click.argument(
    'dir',
    default='.',
    # prompt='Directory to initialize seeqret in',
)
@click.option(
    '--user',
    prompt=True,
    envvar='USERNAME',
    type=str,
    # default=lambda: f"{os.environ['USERDOMAIN']}\\{os.environ['USERNAME']}"
)
@click.option('--email', prompt=True)
@click.option('--pubkey', default=None, show_default=True)
@click.option('--key', default=None, show_default=True)
def init(ctx: click.Context,
         dir: str,
         user: str,
         email: str,
         pubkey: str | None = None,
         key: str | None = None):
    """Initialize a new vault
    """
    dirname = Path(dir).resolve()
    vault_dir = dirname / 'seeqret'

    # we want to create dirname / seeqret

    if not dirname.exists():
        # click.echo(f'The parent of the vault: {dirname} must exist.')
        ctx.fail(f'The parent of the vault: {dirname} must exist.')
        # return

    if not is_writable(dirname):
        # click.echo(f'The parent of the vault: {dirname} must be writable.')
        ctx.fail(f'The parent of the vault: {dirname} must be writable.')
        # return

    if vault_dir.exists():
        if not is_writable(vault_dir):
            ctx.fail(
                f'The vault: {vault_dir} exists and is not writeable, '
                'you must delete it manually.'
            )
            # return
        click.confirm(
            f'The vault: {vault_dir} already exists, overwrite contents?',
            default=True, abort=True)
        # remove_directory(vault_dir)

    seeqret_init.secrets_init(dirname, user, email, pubkey, key)


@cli.command()
@click.pass_context
@click.argument('url')
def fetch(ctx, url):
    "Debugging"
    # XXX: remove me, for debugging...
    seeqret_add.fetch_pubkey_from_url(url)


@cli.group()
def rm():
    """Remove a secret or user from the vault.
    """
    pass


rm.add_command(rm_key)


@cli.group()
def edit():
    """Edit a secret or user in the vault.
    """


def pluralize(items, word, plural):
    return word if len(items) == 1 else plural


@edit.command()
@click.pass_context
@click.argument('filter', default='::')
@click.argument('val')
def value(ctx, filter: str, val: str):
    with cd(os.environ['SEEQRET']):
        storage = SqliteStorage()
        fspec = FilterSpec(filter)
        secrets = storage.fetch_secrets(**fspec.to_filterdict())
        if not secrets:
            ctx.fail(f'No secrets found for {filter}')
        if len(secrets) > 1:
            as_table('app,env,key,value,type', secrets)
            if not click.confirm("Update all values?"):
                ctx.fail('Aborted')
        for secret in secrets:
            secret.value = val
            storage.update_secret(secret)

        click.secho(f"updated {pluralize(secrets, 'secret', 'secrets')}", fg='green')
        as_table('app,env,key,value,type', secrets)


@cli.group()
def add():
    """Add a new secret, key or user
    """
    pass


add.add_command(add_key)


@add.command()
@click.pass_context
@click.option('--username', prompt=True,
              help='Username to record')
@click.option('--email', prompt=True,
              help='Email for the user')
@click.option('--pubkey', prompt=True,
              help='Public key for the user')
def user(ctx, username, email, pubkey):
    """Add a new user to the vault from a public key.

       If the public key is on GitHub, the url is the raw url, e.g.

       https://raw.githubusercontent.com/user/project/refs/heads/main/public.key
    """
    click.secho(f'Adding a new user {username}|{email}|{pubkey}', fg='blue')
    with cd(os.environ['SEEQRET']):
        # click.secho(f'Fetching public key: {url}', fg='blue')
        # pubkey = seeqret_add.fetch_pubkey_from_url(url)
        seeqret_add.add_user(pubkey, username, email)



import click

from seeqret.storage.sqlite_storage import SqliteStorage


# def fetch_pubkey_from_url(url):
#     # FIXME: this does not provide a way to connect the user to the pubkey!
#     r = requests.get(url)
#     if r.status_code != 200:
#         ctx = None
#         try:
#             ctx = click.get_current_context()
#         except AttributeError:
#             # during testing we don't neccessarily have a context...
#             raise RuntimeError(f'Could not fetch pubkey from url: {url}')
#         ctx.fail(click.style(f'Failed to fetch public key: {url}', fg='red'))
#     click.secho('Public key fetched.', fg='green')
#     return r.text


# seeqret add user ...
def add_user(pubkey, username, email):
    click.secho(f'Adding user: {username} with email: {email}', fg='blue')

    storage = SqliteStorage()
    from seeqret.models import User
    users = storage.add_user(User(pubkey=pubkey, username=username, email=email))

    click.secho('User added:', fg='green')
    click.secho(f'    {users[0]}', fg='green')



import os
import re
import sqlite3
from contextlib import contextmanager

from ..models import User, Secret
from .storage import Storage
from logging import getLogger

logger = getLogger(__name__)


def one_line(txt):
    res = txt.replace('\n', ' ').strip()
    return re.sub(r'\s+', ' ', res).strip()


def glob_to_sql(glob_pattern: str) -> str:
    """
    Convert a glob-like pattern (with * and ?) to a SQL WHERE clause string
    and parameters.

    Args:
        glob_pattern (str): Glob pattern to convert.

    Returns:
        tuple: A tuple with a SQL WHERE clause and parameter list.
    """
    sql_pattern = glob_pattern.replace("*", "%").replace("?", "_")
    return sql_pattern


class SqliteStorage(Storage):
    def __init__(self, fname='seeqrets.db'):
        super().__init__('sqlite')
        self.fname = fname

    @contextmanager
    def connection(self):
        path = os.path.join(os.environ['SEEQRET'], self.fname)
        logger.debug('Connecting to SQLite: %s', path)
        # print('Connecting to SQLite: %s' % path)
        cn = sqlite3.connect(path)
        with cn:
            yield cn
        cn.close()
        logger.debug('Closed connection to SQLite: %s', path)

    def execute_sql(self, sql, **filters):
        with self.connection() as cn:
            order_by = ""
            if isinstance(sql, tuple):
                sql, order_by = sql
            params = []
            where = []
            if filters:
                sql += " where "
                for k, v in filters.items():
                    op = 'like' if re.search(r'[\[.*]', v) else '='
                    where.append(f"{k} {op} ? ")
                    params.append(glob_to_sql(v) if op == 'like' else v)
            sql += " and ".join(where) + order_by
            logger.debug('Executing SQL: %s params: %s',
                         one_line(sql), params)
            res = cn.execute(sql, params).fetchall()
            logger.info('Retrieved: %d records', len(res))
            logger.debug('Result: %s', res)
            return res

    def add_user(self, user: User):
        with self.connection() as cn:
            c = cn.cursor()
            c.execute('''
                insert into users (username, email, pubkey)
                values (?, ?, ?);
            ''', (
                user.username,
                user.email,
                user.pubkey
            ))
            cn.commit()

        return self.fetch_users(username=user.username)

    def fetch_users(self, **filters):
        logger.debug('fetch_users: %s', filters)
        sql = ("""
            select username, email, pubkey
            from users
        """, " order by username ")
        return [User(*rec)
                for rec in self.execute_sql(sql, **filters)]

    def update_secret(self, secret: Secret):
        with self.connection() as cn:
            cn.execute('''
                UPDATE secrets SET value = ?
                WHERE app = ? AND env = ? AND key = ?;
            ''', (
                secret._value,
                secret.app, secret.env, secret.key
            ))
            cn.commit()

    def add_secret(self, secret: Secret):
        with self.connection() as cn:
            sql = """
                insert into secrets (app, env, key, value, type)
                values (?, ?, ?, ?, ?);
            """
            cn.execute(sql, (
                secret.app,
                secret.env,
                secret.key,
                secret._value,
                secret.type
            ))
            cn.commit()

    def fetch_secrets(self, **filters):
        logger.debug('fetch_secrets: %s', filters)
        sql = """
            select app, env, key, value, type
            from secrets
        """
        return [Secret(*rec)
                for rec in self.execute_sql(sql, **filters)]

    def remove_secrets(self, **filters):
        sql = """
            delete from secrets
        """
        self.execute_sql(sql, **filters)

    def fetch_admin(self):
        logger.debug('fetch_admin: %s', self)
        with self.connection() as cn:
            admin_rec = cn.execute("""
                select username, email, pubkey
                from users
                where id = 1
            """).fetchone()
        if admin_rec is None:
            return None
        return User(*admin_rec)



from ..seeqrypt.nacl_backend import public_key


class User:
    def __init__(self, username: str, email: str, pubkey: str):
        self.username = username
        self.email = email
        self.pubkey = pubkey

    @property
    def public_key(self):
        return public_key(self.pubkey)

    @property
    def row(self):
        return [self.username, self.email, self.pubkey]

    def __json__(self):
        return dict(
            username=self.username,
            email=self.email,
            pubkey=self.pubkey
        )

    def __str__(self):
        return f'User({self.username}, {self.email}, {self.pubkey})'



"""
Base classes for the serializers.
"""
import click

SERIALIZERS = {}


class ValidationError(ValueError):
    pass


def serializer(cls):
    if not hasattr(cls, 'tag'):
        try:
            ctx = click.get_current_context()
        except RuntimeError:
            raise ValidationError('serializer requires click context')
        ctx.fail('serializer requires click context')
    SERIALIZERS[cls.tag] = cls
    return cls


class BaseSerializer:
    version = 0

    def __init__(self, sender=None, receiver=None,
                 sender_private_key=None, receiver_private_key=None):
        self.sender = sender
        self.receiver = receiver
        self.sender_private_key = sender_private_key
        self.receiver_private_key = receiver_private_key
        self.sender_public_key = sender.public_key
        self.receiver_public_key = receiver.public_key

    def dumps(self, secrets, system):
        raise NotImplementedError()

    def load(self, text):
        raise NotImplementedError()
